/* Generated by Cython 0.29.22 */

#ifndef __PYX_HAVE__py_pal__data_collection__metric
#define __PYX_HAVE__py_pal__data_collection__metric

#include "Python.h"

#ifndef __PYX_HAVE_API__py_pal__data_collection__metric

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C PyObject *__pyx_v_6py_pal_15data_collection_6metric_AVG_BUILTIN_FUNCTION_COMPLEXITY;

#endif /* !__PYX_HAVE_API__py_pal__data_collection__metric */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initmetric(void);
#else
PyMODINIT_FUNC PyInit_metric(void);
#endif

#endif /* !__PYX_HAVE__py_pal__data_collection__metric */
